@@ -52,9 +52,10 @@ from tensorflow.keras import regularizers
 from deep_autoviml.utilities.utilities import print_one_row_from_tf_dataset, print_one_row_from_tf_label
 from deep_autoviml.utilities.utilities import print_classification_metrics, print_regression_model_stats
 from deep_autoviml.utilities.utilities import print_classification_model_stats, plot_history, plot_classification_results
-from deep_autoviml.utilities.utilities import save_valid_predictions
-
+from deep_autoviml.utilities.utilities import save_valid_predictions, print_classification_header
+from deep_autoviml.utilities.utilities import get_callbacks, get_chosen_callback
 from deep_autoviml.modeling.create_model import check_keras_options
+
 #####################################################################################
 from sklearn.metrics import roc_auc_score, mean_squared_error, mean_absolute_error
 from IPython.core.display import Image, display
@@ -82,24 +83,8 @@ def left_subtract(l1,l2):
 import time
 import os
 from sklearn.metrics import balanced_accuracy_score, classification_report, confusion_matrix, roc_auc_score
+import math
 #########################################################################################
-import os
-def get_callbacks(val_mode, val_monitor, patience, save_weights_only):
-    logdir = "deep_autoviml"
-    tensorboard_logpath = os.path.join(logdir,"mylogs")
-    print('    Tensorboard log directory can be found at: %s' %tensorboard_logpath)
-    cp = keras.callbacks.ModelCheckpoint("deep_autoviml", save_best_only=True,
-                                         save_weights_only=save_weights_only, save_format='tf')
-    ### sometimes a model falters and restore_best_weights gives len() not found error. So avoid True option!
-    lr_patience = max(2,int(patience*1.5))
-    rlr = callbacks.ReduceLROnPlateau(monitor=val_monitor, factor=0.25,
-                    patience=lr_patience, min_lr=1e-6, mode='auto', min_delta=0.00001, cooldown=0, verbose=1)
-    es = callbacks.EarlyStopping(monitor=val_monitor, min_delta=0.00001, patience=patience,
-                        verbose=1, mode=val_mode, baseline=None, restore_best_weights=False)
-    tb = callbacks.TensorBoard(tensorboard_logpath)
-    callbacks_list = [cp, rlr, es, tb]
-    return callbacks_list, tensorboard_logpath
-####################################################################################
 ### Split raw_train_set into train and valid data sets first
 ### This is a better way to split a dataset into train and test ####
 ### It does not assume a pre-defined size for the data set.
@@ -136,9 +121,10 @@ def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
     print('    original datasize = %s, initial batchsize = %s' %(data_size, batch_size))
     NUMBER_OF_EPOCHS = check_keras_options(keras_options, "epochs", 100)
     learning_rate = 5e-1
-    steps = max(10, data_size//(batch_size*25))
-    steps = min(40, steps)
+    steps = max(10, 1*(data_size//batch_size))
     print('    recommended steps per epoch = %d' %steps)
+    onecycle_steps = math.ceil(data_size / batch_size) * NUMBER_OF_EPOCHS
+    print('    recommended OneCycle steps = %d' %onecycle_steps)
     STEPS_PER_EPOCH = check_keras_options(keras_options, "steps_per_epoch", 
                         steps)
     optimizer = tf.keras.optimizers.RMSprop(lr=learning_rate)
@@ -155,7 +141,20 @@ def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
     val_monitor = keras_options['monitor']
     val_mode = keras_options['mode']
     patience = keras_options["patience"]
-    callbacks_list, tb_logpath = get_callbacks(val_mode, val_monitor, patience, save_weights_only)
+
+    if keras_options['lr_scheduler'] in ['',"onecycle", "onecycle2"]:
+        #### you need to double the amount of patience for onecycle scheduler ##
+        print('    Increasing the amount of patience for onecycle scheduler')
+        patience = patience * 1.2
+    callbacks_dict, tb_logpath = get_callbacks(val_mode, val_monitor, patience, learning_rate, 
+                            save_weights_only, onecycle_steps)
+    chosen_callback = get_chosen_callback(callbacks_dict, keras_options)
+
+    if keras_options['early_stopping']:
+        callbacks_list = [chosen_callback,  callbacks_dict['tb'],  callbacks_dict['pr'], callbacks_dict['es']]
+    else:
+        callbacks_list = [chosen_callback, callbacks_dict['tb'],  callbacks_dict['pr']]
+
     print('    val mode = %s, val monitor = %s, patience = %s' %(val_mode, val_monitor, patience))
     print('    number of epochs = %d, steps per epoch = %d' %(NUMBER_OF_EPOCHS, STEPS_PER_EPOCH))
     ############## Split train into train and validation datasets here ###############
@@ -172,29 +171,25 @@ def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
     ##################################################################################
     ###   V E R Y    I M P O R T A N T  S T E P   B E F O R E   M O D E L   F I T  ###    
     ##################################################################################
-    #shuffle_size = int(data_size/10)
-    shuffle_size = 100000
+    shuffle_size = int(data_size)
+    #shuffle_size = 100000
     print(' Shuffle size = %d' %shuffle_size)
     train_ds = train_ds.prefetch(batch_size).shuffle(shuffle_size, 
-                            reshuffle_each_iteration=False, seed=42).repeat()
-    valid_ds = valid_ds.prefetch(batch_size).repeat()
+                            reshuffle_each_iteration=False, seed=42)#.repeat()
+    valid_ds = valid_ds.prefetch(batch_size)#.repeat()
     print('Training %s model now. This will take time...' %keras_model_type)
+    
     np.random.seed(42)
     tf.random.set_seed(42)
-    if modeltype == 'Regression':
-        history = deep_model.fit(train_ds, validation_data=valid_ds, 
-                        epochs=NUMBER_OF_EPOCHS, steps_per_epoch=STEPS_PER_EPOCH, 
-                        callbacks=callbacks_list, validation_steps=STEPS_PER_EPOCH,
-                       shuffle=False)
-    else:
-        history = deep_model.fit(train_ds, validation_data=valid_ds, class_weight=class_weights,
-                        epochs=NUMBER_OF_EPOCHS, steps_per_epoch=STEPS_PER_EPOCH, 
-                        callbacks=callbacks_list, validation_steps=STEPS_PER_EPOCH,
-                       shuffle=False)
+    history = deep_model.fit(train_ds, validation_data=valid_ds, class_weight=class_weights,
+                    epochs=NUMBER_OF_EPOCHS, #steps_per_epoch=STEPS_PER_EPOCH, 
+                    callbacks=callbacks_list, #validation_steps=STEPS_PER_EPOCH,
+                   shuffle=False)
 
     print('    Model training metrics available: %s' %history.history.keys())
     try:
-        stopped_epoch = pd.DataFrame(history.history).shape[0] ## this is where it stopped
+        ##### this is where it stopped - you have toi subtract patience from it
+        stopped_epoch = int(pd.DataFrame(history.history).shape[0] - patience )
     except:
         stopped_epoch = 100
 
@@ -249,20 +244,33 @@ def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
         print('    Saved pickle file in %s' %pickle_path)
     except:
         print('Unable to save cat_vocab_dict - please pickle it yourself.')
-
+    
     #################################################################################
     ########     P R E D I C T   O N   H E L D   O U T  D A T A   H E R E      ######
     #################################################################################
-    y_test = np.concatenate(list(heldout_ds.map(lambda x,y: y).as_numpy_iterator()))
+    
+    if num_labels <= 1:
+        y_test = np.concatenate(list(heldout_ds.map(lambda x,y: y).as_numpy_iterator()))
+        print('Single-Label: Heldout data shape: %s' %(y_test.shape,))
+    else:
+        iters = int(data_size/batch_size) + 1
+        for inum, each_target in enumerate(target):
+            add_ls = []
+            for feats, labs in heldout_ds.take(iters): 
+                add_ls.append(list(labs[each_target].numpy()))
+            flat_list = [item for sublist in add_ls for item in sublist]
+            if inum == 0:
+                each_array = np.array(flat_list)
+            else:
+                each_array = np.c_[each_array, np.array(flat_list)]
+        y_test = copy.deepcopy(each_array)
+        print('Multi-Label: Heldout data shape: %s' %(y_test.shape,))
     scores = []
     ls = []
-    print('Held out data actuals shape: %s' %(y_test.shape,))
     if verbose >= 1:
         print_one_row_from_tf_label(heldout_ds)
     ###########################################################################
     y_probas = deep_model.predict(heldout_ds)
-
-    y_test_preds_list = []
     
     if isinstance(target, str):
         if modeltype != 'Regression':
@@ -272,28 +280,23 @@ def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
                 y_test_preds = y_probas.round().astype(int)
             else:
                 y_test_preds = y_probas.ravel()
-        y_test_preds_list.append(y_test_preds)
     else:
         if modeltype != 'Regression':
-            for each_target in target:
-                #### Modify: Not sure about how this will work for multi-class, multi-output problems
-                y_test_preds = y_test_preds.argmax(axis=1)
-            else:
-                if y_test.dtype == 'int':
-                    y_test_preds = y_test_preds.round().astype(int)
+            #### This is for multi-label binary or multi-class problems ##
+            for each_t in range(len(target)):
+                if each_t == 0:
+                    y_test_preds = y_probas[each_t].argmax(axis=1).astype(int)
                 else:
-                    y_test_preds = y_test_preds
-            y_test_preds_list.append(y_test_preds.ravel())
+                    y_test_preds = np.c_[y_test_preds, y_probas[each_t].argmax(axis=1).astype(int)]
         else:
-            ### This is for Multi-Label Regresison ###
-            for each_t in range(len(y_probas)):
+            ### This is for Multi-Label Regression ###
+            for each_t in range(len(target)):
                 if each_t == 0:
                     y_test_preds = y_probas[each_t].mean(axis=1)
                 else:
                     y_test_preds = np.c_[y_test_preds, y_probas[each_t].mean(axis=1)]
-            if y_test.dtype == 'int':
-                y_test_preds = y_test_preds.round().astype(int)
-            y_test_preds_list.append(y_test_preds)
+                if y_test.dtype == 'int':
+                    y_test_preds = y_test_preds.round().astype(int)
 
     print('\nHeld out predictions shape:%s' %(y_test_preds.shape,))
     if verbose >= 1:
@@ -304,7 +307,7 @@ def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
                 print('    Sample predictions: %s' %y_test_preds.ravel()[:10])
             else:
                 print('    Sample predictions:\n%s' %y_test_preds[:10])
-    
+
     #################################################################################
     ########     P L O T T I N G   V A L I D A T I O N   R E S U L T S         ######
     #################################################################################
@@ -312,11 +315,11 @@ def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
     try:
         #print('    Additionally, Tensorboard logs can be found here: %s' %tb_logpath)
         if modeltype == 'Regression':
-            plot_history(history, val_monitor[4:], num_labels)
+            plot_history(history, val_monitor[4:], target)
         elif modeltype == 'Classification':
-            plot_history(history, val_monitor[4:], num_labels)
+            plot_history(history, val_monitor[4:], target)
         else:
-            plot_history(history, val_monitor[4:], num_labels)
+            plot_history(history, val_monitor[4:], target)
     except:
         print('    Plot history is erroring. Tensorboard logs can be found here: %s' %tb_logpath)
 
@@ -325,31 +328,54 @@ def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
     num_labels = cat_vocab_dict['num_labels']
     num_classes = cat_vocab_dict['num_classes']
     if num_labels <= 1:
+        #### This is for Single-Label Problems only ################################
         if modeltype == 'Regression':
-            print_regression_model_stats(y_test, y_test_preds,target,plot_name="deep_autoviml")
+            print_regression_model_stats(y_test, y_test_preds,target,plot_name=project_name)
         else:
-            if num_classes <= 2:
-                labels = target_names = np.unique(y_test)
-                plot_classification_results(y_test, y_probas, labels, target_names)
-                print_classification_metrics(y_test, y_probas, proba_flag=True)
+            print_classification_header(num_classes, num_labels, target)
+            labels = cat_vocab_dict['original_classes']
+            if cat_vocab_dict['target_transformed']:
+                target_names = cat_vocab_dict['transformed_classes']
+                target_le = cat_vocab_dict['target_le']
+                y_pred = y_probas.argmax(axis=1)
+                y_test_trans = target_le.inverse_transform(y_test)
+                y_pred_trans = target_le.inverse_transform(y_pred)
+                plot_classification_results(y_test_trans, y_pred_trans, labels, labels, target)
             else:
-                ###### Use a nice classification matrix printing module here #########
-                labels = target_names = np.unique(y_test)
-                try:
-                    plot_classification_results(y_test, y_probas, labels, target_names)
-                except:
-                    pass
-                print_classification_metrics(y_test, y_probas, proba_flag=True)
-                print(classification_report(y_test,y_test_preds))
-                print(confusion_matrix(y_test, y_test_preds))
+                y_pred = y_probas.argmax(axis=1)
+                plot_classification_results(y_test, y_pred, labels, labels, target)
+            print_classification_metrics(y_test, y_probas, proba_flag=True)
     else:
         if modeltype == 'Regression':
             #### This is for Multi-Label Regression ################################
-            print_regression_model_stats(y_test, y_test_preds,target,plot_name="deep_autoviml")
+            print_regression_model_stats(y_test, y_test_preds,target,plot_name=project_name)
         else:
             #### This is for Multi-Label Classification ################################
-            print_classification_metrics(y_test, y_test_preds, False)
-            print(classification_report(y_test, y_test_preds ))
+            try:
+                targets = cat_vocab_dict["target_variables"]
+                for i, each_target in enumerate(targets):
+                    print_classification_header(num_classes, num_labels, each_target)
+                    labels = cat_vocab_dict[each_target+'_original_classes']
+                    if cat_vocab_dict['target_transformed']:
+                        ###### Use a nice classification matrix printing module here #########
+                        target_names = cat_vocab_dict[each_target+'_transformed_classes']
+                        target_le = cat_vocab_dict['target_le'][i]
+                        y_pred = y_probas[i].argmax(axis=1)
+                        y_test_trans = target_le.inverse_transform(y_test[:,i])
+                        y_pred_trans = target_le.inverse_transform(y_pred)
+                        labels = np.unique(y_test_trans) ### sometimes there is less classes
+                        plot_classification_results(y_test_trans, y_pred_trans, labels, labels, each_target)
+                    else:
+                        y_pred = y_probas[i].argmax(axis=1)
+                        labels = np.unique(y_test[:,i]) ### sometimes there are fewer classes ##
+                        plot_classification_results(y_test[:,i], y_pred, labels, labels, each_target)
+                    print_classification_metrics(y_test[:,i], y_probas[i], proba_flag=True)
+                    #### This prints additional metrics #############
+                    print(classification_report(y_test[:,i],y_test_preds[:,i]))
+                    print(confusion_matrix(y_test[:,i], y_test_preds[:,i]))
+            except:
+                print_classification_metrics(y_test, y_test_preds, False)
+                print(classification_report(y_test, y_test_preds ))
     ### plot the regression results here #########
     if modeltype == 'Regression':
         if isinstance(target, str):
@@ -376,13 +402,9 @@ def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
     ##################################################################################
     ###   V E R Y    I M P O R T A N T  S T E P   B E F O R E   M O D E L   F I T  ###    
     ##################################################################################
-    print('\nTraining full train dataset. This will take time...')
-    full_ds = full_ds.shuffle(shuffle_size).prefetch(batch_size).repeat()
-    if modeltype == 'Regression':
-        deep_model.fit(full_ds, epochs=stopped_epoch, steps_per_epoch=STEPS_PER_EPOCH,
-                 verbose=0)
-    else:
-        deep_model.fit(full_ds, epochs=stopped_epoch, steps_per_epoch=STEPS_PER_EPOCH,
+    print('\nTraining full train dataset for %d epochs. This will take time...' %stopped_epoch)
+    full_ds = full_ds.shuffle(shuffle_size).prefetch(batch_size) #.repeat()
+    deep_model.fit(full_ds, epochs=stopped_epoch, #steps_per_epoch=STEPS_PER_EPOCH,
                  class_weight=class_weights, verbose=0)
 
     print('    completed. Time taken (in mins) = %0.0f' %((time.time()-start_time)/100))
